// ======================================================================
//
// TextureManager_3dfx.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "first.h"
#include "TextureManager_3dfx.h"

#include "Material.h"
#include "MaterialData_3dfx.h"

// ======================================================================

TextureManager::TextureBlock *TextureManager::first;

// ======================================================================

TextureManager::TextureBlock::TextureBlock(int newRegion, u_int32 newLocation, int newLength)
{
	region = newRegion;
	location = newLocation;
	length = newLength;
	userData = NULL;
	previous = NULL;
	next = NULL;
}

// ======================================================================

void TextureManager::install(void)
{
	first = NULL;
}

// ----------------------------------------------------------------------

void TextureManager::remove(void)
{
	TextureBlock *block;

	// delete all the texture blocks
	while (first)
	{
		block = first;
		first = first->next;
		delete block;
	}
}

// ----------------------------------------------------------------------

void TextureManager::addRegion(int region, u_int32 location, int length)
{
	TextureBlock *b;

	// create the texture block
	b = NEW TextureBlock(region, location, length);

	// link the texture block in
	b->next = first;
	if (first)
		first->previous = b;
	first = b;
}

// ----------------------------------------------------------------------

void TextureManager::debugDump(void)
{
}

// ----------------------------------------------------------------------

u_int32 TextureManager::alloc(int size, const MaterialData_3dfx *userData)
{
	TextureBlock *block, *best, *used;

	do
	{
		best = NULL;
		used = NULL;

		for (block = first; block; block = block->next)
			if (block->userData == NULL)
			{
				// pick the smallest unused block that can fit this allocation
				if ((block->length >= size) && (!best || block->length < best->length))
					best = block;
			}
			else
			{
				// pick the LRU block no matter what its size
				if (!used || block->userData->getTagLRU() < used->userData->getTagLRU())
					used = block;
			}

		if (!best)
		{
			// no free texture space was found, so free the LRU texture
			DEBUG_FATAL(!used, ("tmm_3dfx::alloc no texture space"));
			used->userData->unload();
		}

	} while (!best);

	if (best->length > size)
	{
		// subdivide this texture memory block
		block = NEW TextureBlock(best->region, best->location+size, best->length-size);

		// hook the new block in
		block->next = best->next;
		block->previous = best;
		best->next = block;
		best->length = size;
		if (block->next)
			block->next->previous = block;
	}

	// remember information about the material claiming this block
	best->userData = userData;

	return best->location;
}

// ----------------------------------------------------------------------

void TextureManager::free(u_int32 location)
{
	TextureBlock *block, *b;

	for (block = first; block && block->location != location; block = block->next)
		;

	if (!block)
	{
		DEBUG_FATAL(true, ("TextureManager::free not found"));
		return;
	}

	block->userData = NULL;

	// recombine with previous blocks in the same region
	while (block->previous && block->region == block->previous->region && !block->previous->userData)
	{
		// remember this block
		b = block;

		// move to the previous block and update it
		block = block->previous;
		block->length += b->length;
		block->next = b->next;

		// fix the next block's previous
		if (block->next)
			block->next->previous = block;

		delete b;
	}

	// recombine with next blocks in the same region
	while (block->next && block->region == block->next->region && !block->next->userData)
	{
		b = block->next;

		// update the block data
		block->length += b->length;
		block->next = b->next;

		// fix the next block's previous
		if (block->next)
			block->next->previous = block;

		delete b;
	}
}

// ======================================================================
